import numpy as np


class ProteinStructure:
    def __init__(self, name, atom_control, atom_treatment):
        """
        初始化蛋白结构类
        :param name: 蛋白名称
        :param atom_control: 真实原子坐标数组
        :param atom_treatment: 预测原子坐标数组
        """
        self.name = name
        self.atom_control = np.array(atom_control)  # 使用 NumPy 数组
        self.atom_treatment = np.array(atom_treatment)
        self.initial_length = self.length()
        self.distances = None  # 延迟计算

    def length(self):
        """
        计算最大的原子数量
        :return: int: 原子数量
        """
        return max(len(self.atom_control), len(self.atom_treatment))

    def self_check(self):
        """
        自检蛋白原子数量是否一致
        :return: 一个 Bool 值，True 表示自检通过
        """
        return len(self.atom_control) == len(self.atom_treatment)

    def same_len(self):
        """
        修建两组坐标，使它们原子数量强制一致，通过裁剪较长的一部分对齐
        """
        min_len = min(len(self.atom_control), len(self.atom_treatment))
        self.atom_control = self.atom_control[:min_len]
        self.atom_treatment = self.atom_treatment[:min_len]

    def calculate_distances(self):
        """
        强制使原子数量一致后，计算两组原子坐标中每个原子之间的欧几里得距离 (向量化优化)
        :return: 一个存储所有距离的 NumPy 数组
        """
        if len(self.atom_control) != len(self.atom_treatment):
            self.same_len()  # 确保两组坐标长度一致

        # 通过NumPy进行矢量化计算
        differences = self.atom_control - self.atom_treatment  # 每一对坐标点的差值
        distances = np.linalg.norm(differences, axis=1)  # 计算模长
        return distances

    def get_distances(self):
        """
        获取距离，如果尚未计算则执行计算
        """
        if self.distances is None:
            self.distances = self.calculate_distances()  # 延迟计算
        return self.distances


from setuptools import setup, find_packages

setup(
    name="PDB_score",
    version="1.0.1",
    author="SiriNatsume",
    author_email="SiriNatsume@outlook.com",
    description="A tool to massively calculate protein scores using PDB files.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/SiriNatsume/PDB-score",  # 如果有 GitHub 仓库
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "numpy",  # 数值计算
        "biopython",  # 蛋白质结构分析
    ],
    entry_points={
        "console_scripts": [
            "psc=PDB_score.main:main",  # 设置命令行入口
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license = "MIT",
    python_requires=">=3.6",
)


import argparse
import os
from .core import parse_pdb_files, calculate_score
from .ProteinStructure import ProteinStructure


def main():
    parser = argparse.ArgumentParser(description="Calculate protein scores based on PDB files.")
    parser.add_argument("-c", "--control", required=True, help="Experimental PDB file directory")
    parser.add_argument("-t", "--treatment", required=True, help="Predicted PDB file directory")
    parser.add_argument("-o", "--output", required=True, help="Output directory for saving scores")
    parser.add_argument("-T", "--Threat", required=False, type=int, default=4, help="Specify the number of cores, default is 4")
    parser.add_argument("-B", "--Batch", required=False, type= int, default=5000, help="Specify the batch size, default is 5000")
    args = parser.parse_args()

    control_dir = args.control
    treatment_dir = args.treatment
    output_dir = args.output
    thread_count = args.Threat
    batch_size = args.Batch

    # 解析 PDB 文件
    print("Start parsing control PDB files")
    atoms_control = parse_pdb_files(control_dir, batch_size=batch_size, thread_count=thread_count)
    print("Start parsing treatment PDB files")
    atoms_treatment = parse_pdb_files(treatment_dir, batch_size=batch_size, thread_count=thread_count)

    # 为同名文件创建 ProteinStructure 类
    protein_structures = []
    for file_name in atoms_control.keys():
        if file_name in atoms_treatment:
            protein_structure = ProteinStructure(
                name=file_name,
                atom_control=atoms_control[file_name],
                atom_treatment=atoms_treatment[file_name]
            )
            protein_structures.append(protein_structure)


    # 计算评分并保存结果
    # 替代逐行追加列表，用生成器写入 CSV
    with open(os.path.join(output_dir, 'protein_scores.csv'), 'w') as output_csv:
        head = "name," + ",".join(['1A','2A','4A','8A','16A','32A','64A','128A']) + ",Average\n"
        output_csv.write(head)
        for protein in protein_structures:
            scores = {}
            protein.get_distances()  # 获取延迟计算的距离
            for times in range(8):
                limit = 2 ** times
                scores[f'{limit}A'] = calculate_score(protein, limit)
            scores['Average'] = int(sum(scores.values()) / len(scores))

            # 写入数据为行
            line = f"{protein.name}," + ",".join(map(str, scores.values())) + "\n"
            output_csv.write(line)
    print(f"Score is saved to {output_dir}/protein_scores.csv")



if __name__ == "__main__":
    main()


import os
from multiprocessing import Pool
from pathlib import Path

import numpy as np
from Bio.PDB import PDBParser


# 提取中心碳原子坐标
def extract_ca_coordinates(structure):
    ca_coords = [
        residue['CA'].coord
        for model in structure
        for chain in model
        for residue in chain.get_residues()
        if residue.id[0] == ' ' and 'CA' in residue
    ]
    return ca_coords


# 批量解析最多 5000 个 PDB 文件
def parse_pdb_files(directory, batch_size=5000, thread_count=4):
    pdb_path = Path(directory)
    pdb_files = [pdb_file.as_posix() for pdb_file in pdb_path.glob("*.pdb")]
    print(f"Batch size = {batch_size}, Total files = {len(pdb_files)}")
    print(f"Using {thread_count} CPU cores...")


    # 每次处理 batch_size 数量的文件
    all_results = {}
    for i in range(0, len(pdb_files), batch_size):
        batch_files = pdb_files[i:i + batch_size]
        print(f"Processing batch {i // batch_size + 1}: {len(batch_files)} files...")

        # 使用多进程处理每个文件
        with Pool(thread_count) as pool:
            batch_results = pool.map(parse_single_pdb_file, batch_files)

        # 将结果保存到总结果字典
        all_results.update(dict(batch_results))

    return all_results


# 单个文件解析，提取 CA 原子坐标
def parse_single_pdb_file(file_path):
    parser = PDBParser(QUIET=True)
    structure = parser.get_structure(file_path, file_path)
    return os.path.splitext(os.path.basename(file_path))[0], extract_ca_coordinates(structure)




# 优化的 protein 评分计算
def calculate_score(protein, limit):
    limit = float(limit)
    distances = protein.get_distances()

    score = np.sum(distances < limit)
    return int(score / protein.initial_length * 100)